package de.jalin.ldapadmin.ldap;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import org.apache.directory.api.ldap.model.entry.Entry;
import org.apache.directory.api.ldap.model.exception.LdapInvalidDnException;
import org.apache.directory.api.ldap.model.name.Dn;
import org.apache.directory.api.ldap.model.schema.SchemaManager;
import org.apache.directory.server.core.api.DirectoryService;
import org.apache.directory.server.core.api.partition.Partition;
import org.apache.directory.server.core.factory.DefaultDirectoryServiceFactory;
import org.apache.directory.server.core.factory.JdbmPartitionFactory;
import org.apache.directory.server.core.shared.DefaultDnFactory;
import org.apache.directory.server.ldap.LdapServer;
import org.apache.directory.server.protocol.shared.store.LdifFileLoader;
import org.apache.directory.server.protocol.shared.store.LdifLoadFilter;
import org.apache.directory.server.protocol.shared.transport.TcpTransport;

public class DirectoryServiceRunner {

    private static DirectoryServiceRunner serviceRunner = null;
    
    public static void assureServiceRunning(final String name) throws DirectoryServiceException {
        if (serviceRunner == null) {
            final String dnName = "dc=" + name + ",dc=example,dc=com";
            try {
                serviceRunner = new DirectoryServiceRunner("example", dnName, "127.0.0.1", "10389", false, null);
            } catch (Exception e) {
                throw new DirectoryServiceException(e);
            }
        }
    }
    
    public static void main(final String[] args) {
        final String dnString = "dc=" + args[0] + ",dc=example,dc=com";
        final String ip = args[1];
        final String port = args[2];
        try {
            final DirectoryServiceRunner ads = new DirectoryServiceRunner("example", dnString, ip, port, false, null);
            final Entry result = ads.service.getAdminSession().lookup(new Dn(dnString));
            System.out.println("Found entry : " + result);
        } catch (Exception e) {
            System.err.println(e.getMessage());
        }
    }

    private final DirectoryService service;
    private final LdapServer ldapServer;

    public DirectoryServiceRunner(final String partition, final String dnString, final String ip, final String port, final boolean useTLS, final String partitionPath) throws Exception {
        this.service = initService(partition);
        addPartition("ou=config", "config", partitionPath);
        addPartition(dnString, partition, partitionPath);
        this.service.startup();
        loadData();
        this.ldapServer = startServer(ip, port, useTLS);
    }

    private DirectoryService initService(final String partition) throws Exception {
        final DefaultDirectoryServiceFactory factory = new DefaultDirectoryServiceFactory();
        factory.init(partition);
        final DirectoryService directoryService = factory.getDirectoryService();
        directoryService.setShutdownHookEnabled(true);
        directoryService.getChangeLog().setEnabled(false);
        directoryService.setAllowAnonymousAccess(false);
        directoryService.setAccessControlEnabled(true);
        directoryService.setPasswordHidden(false);
        return directoryService;
    }

    private void addPartition(final String dnString, final String partitionId, final String partitionPath) throws LdapInvalidDnException, Exception {
        final JdbmPartitionFactory partitionFactory = new JdbmPartitionFactory();
        final SchemaManager schemaManager = service.getSchemaManager();
        final DefaultDnFactory defaultDnFactory = new DefaultDnFactory(schemaManager, 4096);
        final String parPath = partitionPath != null ? partitionPath : "ldap-data." + Double.valueOf(Math.random()).hashCode();
        final Partition partition = 
            partitionFactory.createPartition(schemaManager, defaultDnFactory, partitionId, dnString, 400, new File(parPath));
        service.addPartition(partition);
    }

    private LdapServer startServer(final String ip, final String port, final boolean useTLS) throws Exception {
        final LdapServer server = new LdapServer();
        final TcpTransport tcpTransport = new TcpTransport(ip, Integer.parseInt(port));
        tcpTransport.enableSSL(useTLS);
        server.setTransports(tcpTransport);
        server.setDirectoryService(service);
        server.start();
        return server;
    }

    public void shutdown() throws Exception {
        if (ldapServer != null) {
            ldapServer.stop();
        }
        if (service != null) {
            service.shutdown();
        }
    }
    
    private void loadData() {
        final File ldifDirectory = new File("ldif");
        if (ldifDirectory.exists() && ldifDirectory.isDirectory()) {
            final File[] ldifFiles;
            ldifFiles = ldifDirectory.listFiles((final File dir, final String name) -> {
                return name.endsWith(".ldif");
            });
            for (final File ldifFile : ldifFiles) {
                final LdifFileLoader ldifFileLoader = new LdifFileLoader(service.getAdminSession(), ldifFile, (List<? extends LdifLoadFilter>) new ArrayList<LdifLoadFilter>());
                ldifFileLoader.execute();
                System.out.println(ldifFile.getName() + " loaded");
            }
        }
    }


    static class DirectoryServiceException extends Exception {

        private static final long serialVersionUID = 1L;

        public DirectoryServiceException(final Throwable exc) {
            super(exc);
        }

    }

}
