package de.jalin.ldapadmin.web;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.Writer;
import java.net.InetAddress;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.attribute.BasicFileAttributes;
import java.nio.file.attribute.FileTime;
import java.security.NoSuchAlgorithmException;
import java.util.Collection;
import java.util.SortedMap;

import javax.servlet.ServletException;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import org.apache.commons.net.smtp.SMTPClient;
import org.apache.commons.net.smtp.SMTPReply;
import org.apache.commons.net.smtp.SimpleSMTPHeader;

import de.jalin.ldapadmin.beans.User;
import de.jalin.ldapadmin.ldap.LDAPSessionException;
import de.jalin.ldapadmin.ldap.SimplePasswordException;
import de.jalin.ldapadmin.ldap.UsersDAO;

@WebServlet(name = "ResetPassword", urlPatterns = {"/passwordreset"}, loadOnStartup = 1)
public class ResetPasswordServlet extends AbstractLDAPServlet {

    private static final long serialVersionUID = 1L;

    private String smtpHost;
    private String smtpPort;
    private String smtpFrom;
    private String tempDir;

    @Override
    public void init() throws ServletException {
        super.init();
        smtpHost = config.getSmtpHost();
        smtpPort = config.getSmtpPort();
        smtpFrom = config.getSmtpFromAddress();
        tempDir = config.getTempDir();
    }

    @Override
    protected void doGet(HttpServletRequest req, HttpServletResponse resp)
            throws ServletException, IOException {
        final HttpSession httpSession = req.getSession();
        cleanSession(httpSession);
        final String token = req.getParameter("token");
        if (token != null && !token.isEmpty()) {
            final String pathToTempFile = tempDir + "/passwd" + token + ".tmp";
			final File passwdResetFile = new File(pathToTempFile);
            final BasicFileAttributes fileAttributes = Files.readAttributes(Paths.get(pathToTempFile), BasicFileAttributes.class);
            final FileTime creationTime = fileAttributes.creationTime();
            final long threeHoursDifference = 10800000L;
			if (creationTime.compareTo(FileTime.fromMillis(System.currentTimeMillis() - threeHoursDifference)) < 0) {
                LOG.warning("password reset token has expired");
                httpSession.setAttribute("errormessage", new Messages(req.getLocale()).getString("ResetPasswordServlet.passwordreset_tokenexpired"));
                req.getRequestDispatcher("/reset-password.jsp").forward(req, resp);
                return;
            }
            if (passwdResetFile.exists() && passwdResetFile.canRead()) {
                try (final BufferedReader reader = new BufferedReader(new FileReader(passwdResetFile))) {
                    final String[] uidAndEMail = reader.readLine().split(":");
                    final UsersDAO usrDAO = new UsersDAO(ldapSession);
                    final User usr = usrDAO.read("uid=" + uidAndEMail[0] + ",ou=users,");
                    httpSession.setAttribute("user", usr);
                    req.getRequestDispatcher("/new-password.jsp").forward(req, resp);
                    return;
                } catch (LDAPSessionException e) {
                    LOG.warning("no valid password reset request");
                    httpSession.setAttribute("errormessage", new Messages(req.getLocale()).getString("ResetPasswordServlet.no_valid_passwordreset_request"));
                    req.getRequestDispatcher("/reset-password.jsp").forward(req, resp);
                    return;
                }
            }
        }
        req.getRequestDispatcher("/reset-password.jsp").forward(req, resp);
    }

    @Override
    protected void doPost(final HttpServletRequest req, final HttpServletResponse resp)
            throws ServletException, IOException {
        final Messages messages = new Messages(req.getLocale());
        final HttpSession httpSession = req.getSession();
        cleanSession(httpSession);
        final UsersDAO usrDAO = new UsersDAO(ldapSession);
        final String loginParam = req.getParameter("login");
        final User sessUsr = (User) httpSession.getAttribute("user");
        if (loginParam != null && sessUsr != null && loginParam.equals(sessUsr.getLogin())) {
            final String password1 = req.getParameter("password");
            final String password2 = req.getParameter("password2");
            if (password1 != null && !password1.isEmpty()) {
                if (password2 == null || !password2.equals(password1)) {
                    httpSession.setAttribute("errormessage", messages.getString("ResetPasswordServlet.passwords_donot_match"));
                    req.getRequestDispatcher("/new-password.jsp").forward(req, resp);
                    return;
                } else {
                    try {
                        sessUsr.setAndValidatePassword(password1);
                        usrDAO.update(sessUsr);
                        httpSession.setAttribute("successmessage", messages.getString("ResetPasswordServlet.password_changed"));
                        httpSession.setAttribute("operation", "profile");
                        httpSession.setAttribute("formdisabled", "disabled");
                        httpSession.setAttribute("iddisabled", "disabled");
                        httpSession.setAttribute("user", sessUsr);
                        req.getRequestDispatcher("/user.jsp").forward(req, resp);
                        return;
                    } catch (SimplePasswordException e) {
                        httpSession.setAttribute("user", sessUsr);
                        httpSession.setAttribute("errormessage", messages.getString("ResetPasswordServlet.simple_password"));
                        req.getRequestDispatcher("/new-password.jsp").forward(req, resp);
                        return;
                    } catch (LDAPSessionException e) {
                        final String excMessage = e.getMessage();
                        if (excMessage != null && excMessage.contains("invalid reuse of password")) { 
                            httpSession.setAttribute("user", sessUsr);
                            httpSession.setAttribute("errormessage", messages.getString("ResetPasswordServlet.invalid_password_reuse"));
                            req.getRequestDispatcher("/new-password.jsp").forward(req, resp);
                            return;
                        }
                        throwServletException(httpSession, e);
                    } catch (NoSuchAlgorithmException e) {
                    	throwServletException(httpSession, e);					}
                }
            }
        }
        final String loginOrEMail = req.getParameter("loginoremail");
        final File tempFile = File.createTempFile("passwd", ".tmp", new File(tempDir));
        try (final PrintStream printStream = new PrintStream(tempFile)) {
            String email = "";
            String login = "";
            String salutation = "";
            if (loginOrEMail != null) {
                if (loginOrEMail.contains("@")) {
                    final SortedMap<String, User> usersMap = usrDAO.loadUsers();
                    final Collection<User> allUsers = usersMap.values();
                    for (User usr : allUsers) {
                        if (usr.getEmail() != null && usr.getEmail().equalsIgnoreCase(loginOrEMail)) {
                            login = usr.getLogin();
                            email = usr.getEmail();
                            salutation = usr.getFirstname() + " " + usr.getLastname();
                        }
                    }
                } else {
                    final User usr = usrDAO.read("uid=" + loginOrEMail + ",ou=users,");
                    if (usr != null) {
                        login = usr.getLogin();
                        email = usr.getEmail();
                        salutation = usr.getFirstname() + " " + usr.getLastname();
                    }
                }
                if (login == null || email == null || login.isEmpty() || email.isEmpty()) {
                    httpSession.setAttribute("errormessage", messages.getString("ResetPasswordServlet.error_sending_password_reset"));
                    req.getRequestDispatcher("/reset-password.jsp").forward(req, resp);
                    return;
                }
                printStream.println(login + ":" + email);
                printStream.close();
            }
            final StringBuffer messageText = new StringBuffer(messages.getString("ResetPasswordServlet.email_greeting"));
            messageText.append(salutation);
            messageText.append(messages.getString("ResetPasswordServlet.email_content"));
            messageText.append(req.getRequestURL().toString().replaceAll("^http\\:", "https://"));
            messageText.append("?token=");
            final String filename = tempFile.getName();
            messageText.append(filename.substring(6, filename.length() - 4));
            messageText.append(messages.getString("ResetPasswordServlet.email_signature"));
            smtpSend(smtpHost, smtpPort, messages, smtpFrom, email, messages.getString("ResetPasswordServlet.email_subject"), messageText.toString());
        } catch (LDAPSessionException | IOException e) {
            LOG.severe("smtp problem");
            httpSession.setAttribute("errormessage", messages.getString("ResetPasswordServlet.error_sending_password_reset"));
            req.getRequestDispatcher("/reset-password.jsp").forward(req, resp);
            return;
        }
        httpSession.invalidate();
        resp.sendRedirect(httpSession.getServletContext().getContextPath() + "/");
    }

    private static void smtpSend(final String smtpHost, final String smtpPort, final Messages messages, final String fromAddress, final String toAddress, final String subject, final String text) throws IOException {
        final SMTPClient client = new SMTPClient();
        final String canonicalHostName = InetAddress.getLocalHost().getHostName();
        client.connect(smtpHost, Integer.parseInt(smtpPort));
        int reply = client.getReplyCode();
        if (!SMTPReply.isPositiveCompletion(reply)) {
            throw new IOException(messages.getString("ResetPasswordServlet.error_sending_email_server"));
        }
        client.login(canonicalHostName);
        client.setSender(fromAddress.trim());
        client.addRecipient(toAddress.trim());
        final Writer sendMessageData = client.sendMessageData();
        if (sendMessageData == null) {
            throw new IOException(messages.getString("ResetPasswordServlet.error_sending_email_relay"));
        }
        try (PrintWriter wr = new PrintWriter(sendMessageData)) {
            final SimpleSMTPHeader header = new SimpleSMTPHeader(fromAddress, toAddress, subject);
            header.addHeaderField("Content-Type", "text/plain; charset=ISO-8859-1");
            header.addHeaderField("Content-Transfer-Encoding", "8bit");
            wr.write(header.toString());
            wr.write(text);
        }
        if (!client.completePendingCommand()) {
            throw new IOException(messages.getString("ResetPasswordServlet.error_sending_email"));
        }
        client.logout();
        client.disconnect();
    }

}
