package de.jalin.ldapadmin.web;

import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;

import javax.servlet.ServletException;
import javax.servlet.annotation.WebServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import de.jalin.ldapadmin.beans.Group;
import de.jalin.ldapadmin.beans.User;
import de.jalin.ldapadmin.beans.ValidationException;
import de.jalin.ldapadmin.ldap.AlreadyBoundException;
import de.jalin.ldapadmin.ldap.GroupsDAO;
import de.jalin.ldapadmin.ldap.LDAPSessionException;
import de.jalin.ldapadmin.ldap.NoGroupMembersException;
import de.jalin.ldapadmin.ldap.RequiredAttributeException;
import de.jalin.ldapadmin.ldap.SimplePasswordException;
import de.jalin.ldapadmin.ldap.UsersDAO;

@WebServlet(name = "LdapUser", urlPatterns = {"/user/*"}, loadOnStartup = 1)
public class UserServlet extends AbstractLDAPServlet {

    private static final long serialVersionUID = 1L;

    @Override
    protected void doGet(final HttpServletRequest req, final HttpServletResponse resp)
            throws ServletException, IOException {
        final String pathInfo = req.getPathInfo();
        String userDN = "";
        if (pathInfo != null && pathInfo.length() > 1 && pathInfo.charAt(0) == '/') {
            userDN = pathInfo.substring(1);
        }
        final HttpSession httpSession = req.getSession();
        cleanSession(httpSession);
        String operation = req.getParameter("op");
        if (operation == null || operation.isEmpty()) {
            operation = "view";
        }
        httpSession.setAttribute("operation", operation);
        httpSession.setAttribute("formdisabled", "view".equals(operation) || "delete".equals(operation) ? "disabled" : "");
        httpSession.setAttribute("iddisabled", "create".equals(operation) ? "" : "disabled");
        final UsersDAO usersDAO = new UsersDAO(ldapSession);
        final GroupsDAO groupsDAO = new GroupsDAO(ldapSession);
        try {
            SortedMap<String, User> users = usersDAO.loadUsers();
            httpSession.setAttribute("users", users);
            httpSession.setAttribute("groups", groupsDAO.loadGroups(users));
            if ("create".equals(operation)) {
                httpSession.setAttribute("user", new User());
            } else {
                httpSession.setAttribute("user", users.get(userDN));
            }
        } catch (LDAPSessionException e) {
            throwServletException(httpSession, e);
        }
        req.getRequestDispatcher("/user.jsp").forward(req, resp);
    }

    @Override
    protected void doPost(HttpServletRequest req, HttpServletResponse resp)
            throws ServletException, IOException {
    	req.setCharacterEncoding("UTF-8");
        final Messages messages = new Messages(req.getLocale());
        final HttpSession httpSession = req.getSession();
        cleanSession(httpSession);
        final String pathInfo = req.getPathInfo();
        String dn = "";
        if (pathInfo != null && pathInfo.length() > 1 && pathInfo.charAt(0) == '/') {
            dn = pathInfo.substring(1);
        }
        String operation = req.getParameter("op");
        if (operation == null || operation.isEmpty()) {
            operation = "none";
        }
        final UsersDAO usersDAO = new UsersDAO(ldapSession);
        final String parLogin = req.getParameter("login");
		final String login = parLogin == null ? null : parLogin.toLowerCase();
        final String firstname = req.getParameter("firstname");
        final String lastname = req.getParameter("lastname");
        final String email = req.getParameter("email");
        final String phone = req.getParameter("phone");
        final String mobile = req.getParameter("mobile");
        final String password = req.getParameter("password");
        final String password2 = req.getParameter("password2");
        final User usr = new User();
        if (!dn.isEmpty()) {
            usr.setDn(dn);
        }
        usr.setLogin(login);
        usr.setFirstname(firstname);
        usr.setLastname(lastname);
        usr.setDisplayname(firstname + " " + lastname);
        usr.setEmail(email);
        usr.setPhone(phone);
        usr.setMobile(mobile);
        final List<String> memberships = new ArrayList<>();
        @SuppressWarnings("unchecked")
        final SortedMap<String, Group> groupsHash = (SortedMap<String, Group>) httpSession.getAttribute("groups");
        final Iterator<String> groupDNIterator = groupsHash.keySet().iterator();
        while (groupDNIterator.hasNext()) {
            final String groupDN = groupDNIterator.next();
            final Group grp = groupsHash.get(groupDN);
            final String isChecked = req.getParameter("check_group_" + grp.getName());
            if (isChecked != null && !isChecked.isEmpty()) {
                memberships.add(grp.getDn());
            }
        }
        usr.setGroups(memberships);
        if (!"delete".equals(operation)) {
            try {
                validatePhone(messages, "phone", phone);
                validatePhone(messages, "mobile", mobile);
                validateEMail(messages, email);
                validateLastName(messages, lastname);
                if (password != null && !password.isEmpty()) {
                    if (password2 == null || !password2.equals(password)) {
                        throw new ValidationException("password2", messages.getString("UserServlet.passwords_donot_match"));
                    } else {
                        usr.setAndValidatePassword(password);
                    }
                }
            } catch (SimplePasswordException e) {
                httpSession.setAttribute("user", usr);
                httpSession.setAttribute("errormessage", messages.getString("ResetPasswordServlet.simple_password"));
                req.getRequestDispatcher("/user.jsp").forward(req, resp);
                return;
            } catch (ValidationException e) {
                httpSession.setAttribute("user", usr);
                httpSession.setAttribute("errormessage", messages.getString("UserServlet.input_field") + " \"" + e.getFieldname() + "\" " + e.getCondition());
                req.getRequestDispatcher("/user.jsp").forward(req, resp);
                return;
            } catch (NoSuchAlgorithmException e) {
                throwServletException(httpSession, e);
			}
        }
        final GroupsDAO groupsDAO = new GroupsDAO(ldapSession);
        try {
            if ("edit".equals(operation)) {
                final User oldValue = usersDAO.read(dn);
                usr.setLogin(oldValue.getLogin());
                usr.setDn(dn);
                usersDAO.update(usr);
                groupsDAO.updateMemberships(usr);
            }
            if ("delete".equals(operation)) {
                final User oldValue = usersDAO.read(dn);
                usr.setLogin(oldValue.getLogin());
                usr.setDn(dn);
                usr.getGroups().clear();
                groupsDAO.updateMemberships(usr);
                usersDAO.delete(usr);
            }
            if ("create".equals(operation)) {
                usr.setLogin(login);
                usersDAO.create(usr);
                groupsDAO.updateMemberships(usr);
                resp.sendRedirect(req.getContextPath() + req.getServletPath() + "/" + usr.getDn());
            } else {
                if ("delete".equals(operation)) {
                    resp.sendRedirect(req.getContextPath() + "/users");
                } else {
                    resp.sendRedirect(req.getContextPath() + req.getServletPath() + pathInfo);
                }
            }
        } catch (LDAPSessionException e) {
            final String excMessage = e.getMessage();
            if (excMessage != null && excMessage.contains("invalid reuse of password")) {
                httpSession.setAttribute("user", usr);
                httpSession.setAttribute("errormessage", messages.getString("UserServlet.invalid_password_reuse"));
                req.getRequestDispatcher("/user.jsp").forward(req, resp);
                return;
            }
            throwServletException(httpSession, e);
        } catch (NoGroupMembersException e) {
            httpSession.setAttribute("user", usr);
            httpSession.setAttribute("errormessage", messages.getString("UserServlet.group_last_member"));
            req.getRequestDispatcher("/user.jsp").forward(req, resp);
        } catch (RequiredAttributeException e) {
            httpSession.setAttribute("user", usr);
            httpSession.setAttribute("errormessage", messages.getString("UserServlet.the_input_field") + " " + e.getFieldname() + " " + messages.getString("UserServlet.is_required"));
            req.getRequestDispatcher("/user.jsp").forward(req, resp);
        } catch (AlreadyBoundException e) {
            httpSession.setAttribute("user", usr);
            httpSession.setAttribute("errormessage", messages.getString("UserServlet.user_exists"));
            req.getRequestDispatcher("/user.jsp").forward(req, resp);
        }
    }

    private void validatePhone(final Messages messages, final String field, final String phone) throws ValidationException {
        if (phone == null || phone.isEmpty()) {
            return;
        }
        final String valid = phone.trim();
        if (!valid.matches("[0-9\\-\\ \\(\\)]*")) {
            throw new ValidationException(field, messages.getString("UserServlet.phone_not_valid"));
        }
    }

    private void validateEMail(final Messages messages, final String email) throws ValidationException {
        if (email == null || email.isEmpty()) {
            return;
        }
        final String valid = email.trim();
        if (!valid.matches("[A-Za-z0-9_+\\.\\-]*@[a-z0-9\\.\\-üöäß]*")) {
            throw new ValidationException("email", messages.getString("UserServlet.email_not_valid"));
        }
    }

    private void validateLastName(final Messages messages, final String lastname) throws ValidationException {
        if (lastname == null || lastname.isEmpty()) {
            throw new ValidationException("lastname", messages.getString("UserServlet.is_required"));
        }
    }

}
